/*
 * Priority Alias (AliasP) directive - always precedes
 * mostly copied from mod_alias.c
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/

#include "apr_strings.h"

#include "httpd.h"
#include "http_log.h"
#include "http_request.h"

#define VERSION "0.1"

module AP_MODULE_DECLARE_DATA aliasp_module;

typedef struct {
    const char *real;
    const char *fake;
} alias_entry;

typedef struct {
    apr_array_header_t *aliases;
} alias_server_conf;

static void *create_config(apr_pool_t *p, server_rec *s)
{
    alias_server_conf *a =
    (alias_server_conf *) apr_pcalloc(p, sizeof(alias_server_conf));

    a->aliases = apr_array_make(p, 20, sizeof(alias_entry));
    return a;
}

static void *merge_config(apr_pool_t *p, void *basev, void *overridesv)
{
    alias_server_conf *a =
    (alias_server_conf *) apr_pcalloc(p, sizeof(alias_server_conf));
    alias_server_conf *base = (alias_server_conf *) basev;
    alias_server_conf *overrides = (alias_server_conf *) overridesv;

    a->aliases = apr_array_append(p, overrides->aliases, base->aliases);
    return a;
}

static const char *add_alias(cmd_parms *cmd, void *dummy, const char *fake,
                             const char *real)
{
    server_rec *s = cmd->server;
    alias_server_conf *conf = ap_get_module_config(s->module_config,
                                                   &aliasp_module);
    alias_entry *new = apr_array_push(conf->aliases);
    new->real = real;
    new->fake = fake;
    return NULL;
}

static const command_rec config_cmds[] =
{
    AP_INIT_TAKE2("AliasP", add_alias, NULL, RSRC_CONF,
                  "a fakename and a realname"),
    {NULL}
};

static int alias_matches(const char *uri, const char *alias_fakename)
{
    const char *aliasp = alias_fakename, *urip = uri;

    while (*aliasp) {
        if (*aliasp == '/') {
            /* any number of '/' in the alias matches any number in
             * the supplied URI, but there must be at least one...
             */
            if (*urip != '/')
                return 0;

            do {
                ++aliasp;
            } while (*aliasp == '/');
            do {
                ++urip;
            } while (*urip == '/');
        }
        else {
            /* Other characters are compared literally */
            if (*urip++ != *aliasp++)
                return 0;
        }
    }

    /* Check last alias path component matched all the way */

    if (aliasp[-1] != '/' && *urip != '\0' && *urip != '/')
        return 0;

    /* Return number of characters from URI which matched (may be
     * greater than length of alias, since we may have matched
     * doubled slashes)
     */

    return urip - uri;
}

static int translate_alias(request_rec *r)
{
    if (r->uri[0] != '/' && r->uri[0] != '\0') {
        return DECLINED;
    }

    ap_conf_vector_t *sconf = r->server->module_config;
    alias_server_conf *serverconf = ap_get_module_config(sconf, &aliasp_module);
    alias_entry *entries = (alias_entry *) serverconf->aliases->elts;
    for (int i = 0; i < serverconf->aliases->nelts; ++i) {
        alias_entry *alias = &entries[i];
        int l = alias_matches(r->uri, alias->fake);

        if (l > 0) {
            ap_set_context_info(r, alias->fake, alias->real);
            char *found = apr_pstrcat(r->pool, alias->real, r->uri + l, NULL);
            r->filename = ap_server_root_relative(r->pool, found);
            return OK;
        }
    }

    return DECLINED;
}

static void register_hooks(apr_pool_t *p)
{
    /* make sure we're called before before mod_proxy and mod_rewrite */
    static const char * const aszSucc[]={ "mod_rewrite.c", "mod_proxy.c", NULL };
    /* filename-to-URI translation */
    ap_hook_translate_name(translate_alias, NULL, aszSucc, APR_HOOK_FIRST);
}

AP_DECLARE_MODULE(aliasp) =
{
    STANDARD20_MODULE_STUFF,
    NULL,                   /* create per-dir config structure */
    NULL,                   /* merge per-dir config structures */
    create_config,          /* create per-server config structure */
    merge_config,           /* merge per-server config structures */
    config_cmds,            /* command-table */
    register_hooks          /* register hooks */
};
